﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using bbiwarg.Utility;

namespace bbiwarg.Recognition.Tracking
{
    public class TrackingStateChangeEventArgs : EventArgs
    {
        public TrackingState NextState { get; private set; }
        public TrackingState PreviousState { get; private set; }

        public TrackingStateChangeEventArgs(TrackingState nextState, TrackingState previousState)
        {
            NextState = nextState;
            PreviousState = previousState;
        }
    }

    public delegate void TrackingStateChangedEventHandler(object sender, TrackingStateChangeEventArgs e);

    class TrackableObjectHistory<T> where T : TrackableObject
    {
        private static List<int> usedIDs = new List<int>();
        public int ID { get; private set; }
        public List<T> TrackedObjects { get; private set; }
        public List<TrackingState> States { get; private set; }
        public T LastObject { get; private set; }
        public T CurrentObject { get { return TrackedObjects[TrackedObjects.Count - 1]; } }
        public TrackingState CurrentState { get { return States[States.Count - 1]; } }
        public int NumFramesInCurrentState { get; private set; }
        private int numFramesDetectedUntilTracked;
        private int numFramesLostUntilDeleted;

        public event TrackingStateChangedEventHandler StateChanged;

        protected virtual void OnStateChanged(TrackingStateChangeEventArgs e) { if (StateChanged != null) StateChanged(this, e); }

        public TrackableObjectHistory(int numFramesDetectedUntilTracked, int numFramesLostUntilDeleted)
        {
            ID = getNextUnusedID();
            TrackedObjects = new List<T>();
            TrackedObjects.Add(null);
            States = new List<TrackingState>();
            States.Add(TrackingState.None);
            NumFramesInCurrentState = 1;
            this.numFramesDetectedUntilTracked = numFramesDetectedUntilTracked;
            this.numFramesLostUntilDeleted = numFramesLostUntilDeleted;
        }

        ~TrackableObjectHistory()
        {
            setIDUnused(ID);
        }

        public void addObjectToHistory(T detectedObject)
        {
            TrackingState previousState = CurrentState;
            TrackingState newState = TrackingState.None;

            //get newState
            if (detectedObject != null)
            {
                if (previousState == TrackingState.None)
                    newState = TrackingState.Detected;
                else if (previousState == TrackingState.Lost)
                    newState = TrackingState.Tracked;
                else if (previousState == TrackingState.Tracked)
                    newState = TrackingState.Tracked;
                else if (previousState == TrackingState.Detected)
                {
                    if (NumFramesInCurrentState == numFramesDetectedUntilTracked)
                        newState = TrackingState.Tracked;
                    else
                        newState = TrackingState.Detected;
                }
            }
            else if (previousState == TrackingState.Lost && NumFramesInCurrentState == numFramesLostUntilDeleted)
                newState = TrackingState.Delete;
            else
                newState = TrackingState.Lost;

            //update numFramesInCurrentState
            if (newState == previousState)
                NumFramesInCurrentState++;
            else
                NumFramesInCurrentState = 1;

            //update lastObject
            if (detectedObject != null)
            {
                LastObject = detectedObject;
                detectedObject.setTrackID(ID);
            }

            TrackedObjects.Add(detectedObject);
            States.Add(newState);
            OnStateChanged(new TrackingStateChangeEventArgs(newState, previousState));
        }

        public static int getNextUnusedID()
        {
            int id = 1;
            while (usedIDs.Contains(id))
                id++;
            usedIDs.Add(id);
            return id;
        }

        public static void setIDUnused(int id)
        {
            usedIDs.Remove(id);
        }
    }
}
